/******************************************************************************
 * Boot Loader v1.0                                                           *
 * di Valerio Di Gregorio                                                     *
 *                                                                            *
 * Boot Loader v1.0 carica da shell di FreeDOS il "Nucleo di Calcolatori      *
 * Elettronici v1.0". Questo progetto  conforme alle "Multiboot              *
 * Specification version 0.6.95".                                             *
 ******************************************************************************/

#include <stdlib.h>
#include <string.h>
#include "dos.h"
#include "coff16.h"

int initialize_exe(FILE* obj, ExeCoff_go32* exe) {
	
    struct DOS_EXE dos;                                             /* Struttura iniziale prevista da DOS */

    /* Inizializzazione */
    exe->obj = obj;
    exe->seghdr = NULL;

    /* Spostamento all'inizio del file */
	if (fseek(exe->obj, 0, SEEK_SET) != 0)
		return 0;
    
    /* Lettura della struttura iniziale */
	if (fread(&dos, sizeof(struct DOS_EXE), 1, exe->obj) < 1)
		return 0;

    /* Controllo del numero magico */
	if (dos.signature != DOS_MAGIC)
		return 0;

    /* Calcolo del punto di inizio */
	exe->soff = dos.blocks_in_file * 512L;
	if (dos.bytes_in_last_block)
		  exe->soff -= (512 - dos.bytes_in_last_block);

    /* Spostamento all'inizio dell'oggetto COFF */
	if (fseek(exe->obj, exe->soff, SEEK_SET) != 0)
		return 0;

    /* Lettura del file header */
	if (fread(&exe->filhdr, FILHSZ, 1, exe->obj) < 1)
		return 0;
		
    /* Controllo del numero magico */
	if (exe->filhdr.f_magic != I386MAGIC)
		return 0;

    /* Controllo dei flags */
	if (!(exe->filhdr.f_flags & F_EXEC))
		return 0;

    /* Lettura dell'optional header */
	if (fread(&exe->opthdr, 1, exe->filhdr.f_opthdr, exe->obj) < exe->filhdr.f_opthdr) {
		fprintf(stderr, "Fine prematura del file COFF-go32\n");
		exit(EXIT_FAILURE);
	}

	/* Controllo del numero magico */
	if (exe->opthdr.magic != ZMAGIC)
		return 0;

	/* Lettura dei section headers */
	if ((exe->seghdr = (SCNHDR*)calloc(exe->filhdr.f_nscns, SCNHSZ))==NULL) {
        fprintf(stderr, "Memoria insufficiente\n");
	    exit(EXIT_FAILURE);
    }

	if (fread(exe->seghdr, SCNHSZ, exe->filhdr.f_nscns, exe->obj) < exe->filhdr.f_nscns) {
		fprintf(stderr, "Fine prematura del file COFF-go32\n");
		exit(EXIT_FAILURE);
	}

	return 1;
}

void destroy_exe(ExeCoff_go32* exe) {    
     free(exe->seghdr);    
}

void initialize_img(ImgCoff_go32* img, unsigned long n) {    
     
     int i;
   
     /* Inizializzazione dell'immagine */
	 img->n = n;
	 img->ds = get_ds();
	 img->segment.ul = 0;
	 img->size.ul = 0;
	 img->vaddr.ul = 0;
	 img->entry_point.ul = 0;
	 
	 if ((img->segment.v = (pointer*)calloc(img->n, sizeof(pointer)))==NULL) {
         fprintf(stderr, "Memoria insufficiente\n");
	     exit(EXIT_FAILURE);
     }
 	 if ((img->size.v = (unsigned long*)calloc(img->n, sizeof(unsigned long)))==NULL) {
         fprintf(stderr, "Memoria insufficiente\n");
	     exit(EXIT_FAILURE);
     }
     if ((img->vaddr.v = (unsigned long*)calloc(img->n, sizeof(unsigned long)))==NULL) {
         fprintf(stderr, "Memoria insufficiente\n");
	     exit(EXIT_FAILURE);
     }
  
     for(i=0; i<img->n; i++) {
         img->segment.v[i].ul = 0;
	     img->size.v[i] = 0;
	     img->vaddr.v[i] = 0;
     }
     
}

void destroy_img(ImgCoff_go32* img) {
     int i;
      
     for(i=0; i<img->n; i++)
         free(img->segment.v[i].v);

	 free(img->segment.v);
 	 free(img->size.v);
     free(img->vaddr.v);
     
}

void load_coff(char* path, ImgCoff_go32* img){

   FILE* obj = NULL;              /* File formato COFF */
   ExeCoff_go32 exe;              /* Informazioni sull'eseguibile COFF */
   int loaded = 0;                /* Contatore di segmenti caricati */
   int skipped = 0;               /* Contatore di segmenti non caricati */
   int disp;                      /* Spiazzamento dell'entry point all'interno del segmento */
   int i;

   /* Apertura file */
   if(!(obj = fopen(path, "r+b"))) {
       	perror(path);
	   exit(EXIT_FAILURE);
   }

   /* Interpretazione dell'eseguibile COFF */
   if(!initialize_exe(obj, &exe)) {
       fprintf(stderr, "File `%s' non eseguibile o non conforme al formato COFF-go32\n", path);
	   exit(EXIT_FAILURE);
   }

   /* Caricamento dei segmenti */
   initialize_img(img, exe.filhdr.f_nscns);

   for(i=0; i<exe.filhdr.f_nscns; i = loaded + skipped) {
       
       /* Salto di tutti i segmenti salvo quelli TEXT, DATA e BSS */       
       if (exe.seghdr[i].s_vaddr == 0 ||
		   exe.seghdr[i].s_flags != STYP_TEXT &&
           exe.seghdr[i].s_flags != STYP_DATA &&
           exe.seghdr[i].s_flags != STYP_BSS) {
           
           skipped++;
           continue;
       }

       /* Salto di segmenti vuoti */
       if (exe.seghdr[i].s_size <= 0) {
           skipped++;
           continue;
       }  

       /* Allocazione dello spazio necessario per il contenuto del segmento */
       if (exe.seghdr[i].s_flags & STYP_BSS) {
		img->segment.v[loaded].v = 0;
       } else {
	       if((img->segment.v[loaded].v = (byte*)malloc(exe.seghdr[i].s_size))==0) {
		   fprintf(stderr, "Memoria insufficiente\n");
		   exit(EXIT_FAILURE);
	       }
       }

       /* Inizializzazione del segmento */
       img->size.v[loaded] = exe.seghdr[i].s_size;
       img->vaddr.v[loaded] = exe.seghdr[i].s_vaddr;

       /* Posizionamento dell'entry point */       
       disp = exe.opthdr.entry - exe.seghdr[i].s_vaddr;
       
       if (exe.seghdr[i].s_flags == STYP_TEXT &&
           disp >= 0 && disp < exe.seghdr[i].s_size) {

           img->entry_point.ul = img->segment.v[loaded].ul + disp;
       }
       
       /* Spostamento sul offset relativo al contenuto del segmento */
       if (fseek(exe.obj, exe.soff + exe.seghdr[i].s_scnptr, SEEK_SET) != 0) {
		   perror(path);
		   exit(EXIT_FAILURE);
	   }

       /* Caricamento del contenuto del segmento */
	   if (!(exe.seghdr[i].s_flags & STYP_BSS)) {
		   if (fread(img->segment.v[loaded].v, 1, exe.seghdr[i].s_size, exe.obj) < exe.seghdr[i].s_size) {
			   perror(path);
			   exit(EXIT_FAILURE);
		   }
	   }

       loaded++;   
 
   }

   /* Aggiornamento con l'effettivo numero di segmenti caricati */
   img->n = loaded;     /* "skipped" elementi degli array rimangono
                           inutilizzati (si potrebbe usare una realloc) */
   
   /* Distruzione della struttura di informazioni sull'eseguibile COFF e chiusura del file */
   destroy_exe(&exe);
   if (fclose(obj)){
	   perror(path);
	   exit(EXIT_FAILURE);
   }
}

/* Converte un indirizzo espesso come selettore:offset in indirizzo lineare */
unsigned long vaddr(unsigned int segment, unsigned int offset) {
   unsigned long tmp = segment;
   tmp *= 16;
   tmp += offset;
   return tmp;           /* Selettore*16 + Offset */
}
